#!/usr/bin/env python3

from __future__ import with_statement

from abc import ABC, abstractmethod
import os
import logging
import sys
import errno
import functools
import re

from fuse import FUSE, FuseOSError, Operations, LoggingMixIn

logger = logging.getLogger('fuse.log-mixin')
logger.setLevel(logging.DEBUG)

from .handler import *

class Router:
    """
    A router that matches filesystem operations at a given path
    to a handler
    """
    def __init__(self, client):
        self.client = client
        self.handlers = [
            (r"^/$", BaseHandler(client)),
            (r"^/user$", UserHandler(client)),
            (r"^/teams\/?$", TeamsHandler(client)),
            (r"^/teams/[0-9]+$", TeamHandler(client))

        ]

    def match(self, operation: str, path: str):
        """
        Determine if a handler existers for the operation
        at a given path
        """
        for reg, handler in self.handlers:
            if re.match(reg, path):
                print("Found matching handler for", operation, path)
                method = getattr(handler, operation)
                return method
        raise Exception(f"No handler for {operation} at {path}")
 
def delegate(func):
    """
    Decorator that delegates execution of a FUSE callback to
    the appropriate handler.
    """
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        path = args[0]
        handler = self.router.match(func.__name__, path)
        return handler(*args, **kwargs)
    return wrapped

class ClickupFS(LoggingMixIn):
    """A FUSE filesystem for Clickup"""
    def __init__(self, root: str, clickup_client):
        self.root = root
        self.client = clickup_client
        self.router = Router(self.client)

    # Filesystem methods

    @log
    @delegate
    def access(self, path: str, mode: int) -> None:
        pass

    @log
    @delegate
    def chmod(self, path: str, mode) -> None:
        pass

    @log
    @delegate
    def chown(self, path: str, uid, gid) -> None:
        pass

    @log
    @delegate
    def getattr(self, path:str, fh=None) -> dict:
        pass

    @log
    @delegate
    def readdir(self, path: str, fh: int) -> list:
        pass

    @log
    @delegate
    def readlink(self, path: str):
        pass

    @log
    @delegate
    def mknod(self, path, mode, dev):
        pass

    @log
    @delegate
    def rmdir(self, path: str):
        pass

    @log
    @delegate
    def mkdir(self, path, mode):
        pass

    @log
    @delegate
    def statfs(self, path):
        #https://linux.die.net/man/2/statvfs
        pass

    @log
    @delegate
    def unlink(self, path):
        pass

    @log
    @delegate
    def symlink(self, name, target):
        pass

    @log
    @delegate
    def rename(self, old, new):
        pass

    @log
    @delegate
    def link(self, target, name):
        pass

    @log
    @delegate
    def utimens(self, path, times=None):
        pass

    # File methods
    @log
    @delegate
    def open(self, path, flags):
        pass

    @log
    @delegate
    def create(self, path, mode, fi=None):
        pass

    @log
    @delegate
    def read(self, path, length, offset, fh):
        pass

    @log
    @delegate
    def write(self, path, buf, offset, fh):
        pass

    @log
    @delegate
    def truncate(self, path, length, fh=None):
        pass

    @log
    @delegate
    def flush(self, path, fh):
        pass

    @log
    @delegate
    def release(self, path, fh):
        pass

    @log
    @delegate
    def fsync(self, path, fdatasync, fh):
        pass

def clickup_fuse(mountpoint: str, clickup_client):
    """Create a fuse filesystem for the clickup api"""
    return FUSE(ClickupFS(mountpoint, clickup_client), mountpoint, nothreads=True, foreground=True)




import functools
import json
import os

def log(func):
    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        print(f"{func.__name__.upper()} with {args}, {kwargs}")
        ret = func(self, *args, **kwargs)
        print("Returned", ret)
        return ret
    return inner

class Handler:
    """A base handler for FUSE operations"""
    def __init__(self, clickup_client):
        self.client = clickup_client

    def access(self, path: str, mode: int) -> None:
        return None

    def chmod(self, path: str, mode) -> None:
        return None

    def chown(self, path: str, uid: str, gid: str) -> None:
        return None

    def getattr(self, path:str, fh=None) -> dict:
        pass

    def readdir(self, path: str, fh: int) -> list:
        pass

    def readlink(self, path: str):
        pass

    def mknod(self, path, mode, dev):
        pass

    def rmdir(self, path):
        pass

    def mkdir(self, path, mode):
        pass

    def statfs(self, path: str):
        return {
            "f_bsize": 1048576,
            "f_frsize": 4096,
            "f_blocks": 1220613,
            "f_bfree": 760260,
            "f_bavail": 750614,
            "f_files": 42949672,
            "f_ffree": 42928496,
            "f_favail": 42928496,
            "f_flag": 0,
            "f_namemax": 255
        }

    def unlink(self, path: str):
        pass

    def symlink(self, name, target):
        pass

    def rename(self, old: str, new: str):
        pass

    def link(self, target, name):
        pass

    def utimens(self, path: str, times=None):
        pass

    # File methods
    # ============

    def open(self, path: str, flags):
        pass

    def create(self, path: str, mode, fi=None):
        pass

    def read(self, path: str, length: int, offset: int, fh):
        pass

    def write(self, path: str, buf, offset: str, fh):
        pass

    def truncate(self, path:str , length, fh=None):
        pass

    def flush(self, path: str, fh):
        pass

    def release(self, path: str, fh):
        pass

    def fsync(self, path: str, fdatasync, fh):
        pass

class BaseHandler(Handler):
    """Handler for operations at the root of the FUSE filesystem"""
    @log
    def getattr(self, path:str, fh=None) -> dict:
        return {
            # time of last access
            "st_atime": 1557351945,
            # Time of last status change
            "st_ctime": 1557351945,
            # time of last modification
            "st_mtime": 1557351945,
            # Group that owns the file
            "st_gid": os.getgid(),
            # File or folder
            "st_mode":  16877,
            # Number of hard links
            "st_nlink": 1,
            # Size of file in bytes
            "st_size": 4096,
            # user id
            "st_uid": os.getuid()
        }

    @log
    def readdir(self, path: str, fh: int = None) -> list:
        return (".", "..", "user", "teams")


class UserHandler(Handler):
    """Handler for retrieving a user"""
    @log
    def getattr(self, path: str, fh: int = None) -> dict:
        user = self.client.user()
        data = bytes(json.dumps(user,indent=2), "utf8")
        return {
            "st_atime": 1557351945,
            "st_ctime": 1557351945,
            "st_mtime": 1557351945,
            "st_gid": os.getgid(),
            "st_mode": 33188,
            "st_nlink": 1,
            "st_size": len(data),
            "st_uid": os.getuid()
        }

    @log
    def open(self, path: str, flags):
        return 0

    @log
    def read(self, path: str, length: int, offset: int, fh: int):
        user = self.client.user()
        data = bytes(json.dumps(user,indent=2), "utf8")
        return data[offset:offset+length]

class TeamsHandler(Handler):
    """Handler for retrieving a list of teams"""
    @log
    def getattr(self, path: str, fh:int = None) -> dict:
        return {
            "st_atime": 1557351945,
            "st_ctime": 1557351945,
            "st_mtime": 1557351945,
            "st_gid": os.getgid(),
            "st_mode": 16877,
            "st_nlink": 1,
            "st_size": 1024,
            "st_uid": os.getuid()
        }

    @log
    def readdir(self, path: str, fh: int = None) -> list:
        teams = self.client.teams()["teams"]
        ids = [team["id"] for team in teams]
        return [".", ".."] + ids

class TeamHandler(Handler):
    """Handler for retrieving a team"""
    @log
    def getattr(self, path: str, fh: int = None) -> dict:
        _id = path.split("/")[-1]
        team = self.client.team(_id)
        data = bytes(json.dumps(team,indent=2), "utf8")
        return {
            "st_atime": 1557351945,
            "st_ctime": 1557351945,
            "st_mtime": 1557351945,
            "st_gid": os.getgid(),
            "st_mode": 33188,
            "st_nlink": 1,
            "st_size": len(data),
            "st_uid": os.getuid()
        }

    @log
    def open(self, path: str, flags):
        return 0

    @log
    def read(self, path: str, length: int, offset: int, fh: int):
        _id = path.split("/")[-1]
        team = self.client.team(_id)
        data = bytes(json.dumps(team, indent=2), "utf8")
        return data[offset:offset+length]



import os
from setuptools import setup

# Taken from https://github.com/kennethreitz/setup.py/blob/master/setup.py
here = os.path.abspath(os.path.dirname(__file__))

setup(
    name='clickupy',
    version='0.0.3',
    description='Package for interacting with the clickup api',
    url='https://github.com/dang3r/clickupy',
    author='Daniel Cardoza',
    author_email='dan@danielcardoza.com',
    license='MIT',
    packages=['clickupy'],
    python_requires='>=3.4.0',
    scripts=['bin/clickup'],
    install_requires=['click', 'requests'],
    long_description='\n' + open(os.path.join(here, 'README.md')).read(),
    long_description_content_type='text/markdown',
    zip_safe=False
)



#!/usr/bin/env python3

from collections import defaultdict
from functools import update_wrapper
import json

import requests

from .fuse import *

class Error(Exception):
    pass

class SerializerDoesNotExist(Error):
    pass

class Client:
    """An API client for click"""
    def __init__(self, api_key: str, url="https://api.clickup.com"):
        self.api_key = api_key
        self.url = url + "/api/v1/"
        self.session = requests.Session()
        self.session.headers.update({"Authorization": api_key})

    def _do(self, method, route, params=None, data=None):
        resp = self.session.request(
            method,
            self.url + route,
            params=params,
            data=data
        )
        resp.raise_for_status()
        return resp.json()

    def user(self):
        return self._do("GET", "user")

    def teams(self):
        return self._do("GET", "team")

    def team(self, id: str):
        return self._do("GET", f"team/{id}")

    def spaces(self, id: str):
        return self._do("GET", f"team/{id}/space")

    def projects(self, id: str):
        return self._do("GET", f"space/{id}/project")

    def create_list(self, project_id: str, list_name:str):
        data = {"name": list_name}
        return self._do("POST", f"project/{project_id}/list", data=data)

    def edit_list(self, project_id: str, list_name: str):
        data = {"name": list_name}
        return self._do("PUT", f"project/{project_id}/list", data=data)

    def tasks(self, team_id: str, **kwargs):
        """See documentation for all kwargs"""
        return self._do("GET", f"team/{team_id}/task", params=kwargs)

    def create_task(self, list_id:str, name:str, content:str, assignees:list, status:str, priority:int, due_date:str):
       data = {
           "name": name,
           "content": content,
           "assignees": assignees,
           "status": status,
           "priority": priority,
           "due_date": due_date
       }
       return self._do("POST", f"list/{list_id}/task")

    def edit_task(self, list_id:str, name:str, content:str, assignees:list, status:str, priority:int, due_date:str):
       data = {
           "name": name,
           "content": content,
           "assignees": assignees,
           "status": status,
           "priority": priority,
           "due_date": due_date
       }
       return self._do("PUT", f"list/{list_id}/task")

class SerializerFactory:
    def __init__(self):
        self.serializers = defaultdict(dict)

    def __call__(self, object_type: str, format: str):
        def wrapped(func):
            self.serializers[object_type][format] = func
            return func
        return wrapped

    def __contains__(self, key: tuple):
        object_type, format = key
        return object_type in self.serializers and format in self.serializers[object_type]

serializer = SerializerFactory()

@serializer("user", "human")
def user_human(user: dict):
    user = user["user"]
    return f'{user["id"]} {user["username"]}'

@serializer("team", "human")
def team_human(team: dict):
    return f'{team["id"]} {team["name"]} {len(team["members"])} users'

@serializer("teams", "human")
def teams_human(teams: list):
    teams_str = [team_human(team) for team in teams]
    return "\n".join(teams_str)

@serializer("spaces", "human")
def spaces_human(spaces: list):
    spaces_str = [f'{space["id"]} {space["name"]}' for space in spaces]
    return "\n".join(spaces_str)

@serializer("projects", "human")
def projects_human(projects: list):
    projects_str = [f'{project["id"]} {project["name"]}' for project in projects]
    return "\n".join(projects_str)

@serializer("tasks", "human")
def tasks_human(tasks: list):
    tasks_str = [f'{task["id"]} {task["name"]}' for task in tasks]
    return "\n".join(tasks_str)

def serialize(object_type: str, object: dict, format: str):
    if (object_type, format) not in serializer:
        raise SerializerDoesNotExist(object_type, format)
    elif object_type == "json":
        return json.dumps(object, indent=2)
    return serializer.serializers[object_type][format](object)
